# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.constants import EIEIO_COMMAND_IDS
from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader

_PATTERN_BBBxBBI = struct.Struct("<BBBxBBI")


class SpinnakerRequestBuffers(EIEIOCommandMessage):
    """
    Message used in the context of the buffering input mechanism which is
    sent by the SpiNNaker system to the host computer to ask for more data
    to inject during the simulation.
    """
    __slots__ = (
        "_region_id",
        "_sequence_no",
        "_space_available",
        "_p", "_x", "_y")

    def __init__(
            self, x: int, y: int, p: int, region_id: int, sequence_no: int,
            space_available: int):
        """

        :param int x:
        :param int y:
        :param int p:
        :param int region_id:
        :param int sequence_no:
        :param int space_available:
        """
        # pylint: disable=too-many-arguments
        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.SPINNAKER_REQUEST_BUFFERS))
        self._x = x
        self._y = y
        self._p = p
        self._region_id = region_id
        self._sequence_no = sequence_no
        self._space_available = space_available

    @property
    def x(self) -> int:
        """
        Gets the x value passed into the init

        :rtype: int
        """
        return self._x

    @property
    def y(self) -> int:
        """
        Gets the y value passed into the init

        :rtype: int
        """
        return self._y

    @property
    def p(self) -> int:
        """
        Gets the p value passed into the init

        :rtype: int
        """
        return self._p

    @property
    def region_id(self) -> int:
        """
        Gets the region_id value passed into the init

        :rtype: int
        """
        return self._region_id

    @property
    def sequence_no(self) -> int:
        """
        Gets the sequence_no value passed into the init

        :rtype: int
        """
        return self._sequence_no

    @property
    def space_available(self) -> int:
        """
        Gets the space_available value passed into the init

        :rtype: int
        """
        return self._space_available

    @staticmethod
    @overrides(EIEIOCommandMessage.get_min_packet_length)
    def get_min_packet_length() -> int:
        return 12

    @staticmethod
    @overrides(EIEIOCommandMessage.from_bytestring)
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "SpinnakerRequestBuffers":
        y, x, processor, region_id, sequence_no, space = \
            _PATTERN_BBBxBBI.unpack_from(data, offset)
        p = (processor >> 3) & 0x1F
        return SpinnakerRequestBuffers(
            x, y, p, region_id & 0xF, sequence_no, space)

    @property
    @overrides(EIEIOCommandMessage.bytestring)
    def bytestring(self) -> bytes:
        return (super().bytestring + _PATTERN_BBBxBBI.pack(
            self._y, self._x, self._p << 3, self._region_id,
            self._sequence_no, self._space_available))



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional, Tuple
from spinn_utilities.overrides import overrides
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.messages.scp.enums import SCPResult
from spinnman.connections.abstract_classes import AbstractSCPConnection
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from .sdp_connection import SDPConnection

_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP = struct.Struct("<2x")


class SCAMPConnection(SDPConnection, AbstractSCPConnection):
    """
    A UDP connection to SCAMP on the board.
    """
    __slots__ = ()

    def __init__(
            self, chip_x: int = 255, chip_y: int = 255,
            local_host: Optional[str] = None,
            local_port: Optional[int] = None,
            remote_host: Optional[str] = None,
            remote_port: Optional[int] = None):
        """
        :param int chip_x:
            The x-coordinate of the chip on the board with this remote_host
        :param int chip_y:
            The y-coordinate of the chip on the board with this remote_host
        :param str local_host: The optional IP address or host name of the
            local interface to listen on
        :param int local_port: The optional local port to listen on
        :param str remote_host: The optional remote host name or IP address to
            send messages to.  If not specified, sending will not be possible
            using this connection
        :param int remote_port: The optional remote port number to send
            messages to. If not specified, sending will not be possible using
            this connection
        """
        # pylint: disable=too-many-arguments
        if remote_port is None:
            remote_port = SCP_SCAMP_PORT
        super().__init__(
            chip_x, chip_y, local_host, local_port, remote_host, remote_port)

    @property
    @overrides(AbstractSCPConnection.chip_x)
    def chip_x(self) -> int:
        return self._chip_x

    @property
    @overrides(AbstractSCPConnection.chip_y)
    def chip_y(self) -> int:
        return self._chip_y

    def update_chip_coordinates(self, x: int, y: int):
        """
        Sets the coordinates without checking they are valid.

        :param int x:
        :param int y:
        """
        self._chip_x = x
        self._chip_y = y

    @overrides(AbstractSCPConnection.get_scp_data,
               additional_arguments=['x', 'y'], extend_defaults=True)
    def get_scp_data(
            self, scp_request: AbstractSCPRequest,
            x: Optional[int] = None, y: Optional[int] = None) -> bytes:
        """
        :param int x: Optional: x-coordinate of where to send to
        :param int y: Optional: y-coordinate of where to send to
        """
        # pylint: disable=arguments-differ
        if x is None:
            x = self.chip_x
        if y is None:
            y = self.chip_y
        scp_request.sdp_header.update_for_send(x, y)
        return _TWO_SKIP.pack() + scp_request.bytestring

    @overrides(AbstractSCPConnection.receive_scp_response)
    def receive_scp_response(self, timeout: Optional[float] = 1.0) -> Tuple[
            SCPResult, int, bytes, int]:
        data = self.receive(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2

    def receive_scp_response_with_address(
            self, timeout: float = 1.0) -> Tuple[
                SCPResult, int, bytes, int, str, int]:
        """

        :param float timeout:
        :rtype: tuple(SCPResult, int, bytes, int, str, int)
        """
        data, (addr, port) = self.receive_with_address(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2, addr, port

    def __repr__(self) -> str:
        return (
            f"SCAMPConnection(chip_x={self._chip_x}, chip_y={self._chip_y}, "
            f"local_host={self.local_ip_address}, local_port={self.local_port}"
            f", remote_host={self.remote_ip_address}, "
            f"remote_port={self.remote_port})")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class CPUState(Enum):
    """
    SARK CPU States.
    """
    #: Dead core
    DEAD = 0
    #: Powered down
    POWERED_DOWN = 1
    #: Died with Run Time Error
    RUN_TIME_EXCEPTION = 2
    #: Watchdog expired
    WATCHDOG = 3
    #: Initialising (transient)
    INITIALISING = 4
    #: Ready to execute
    READY = 5
    #: Entered c_main
    C_MAIN = 6
    #: Running (API/Event)
    RUNNING = 7
    #: Waiting for sync 0
    SYNC0 = 8
    #: Waiting for sync 1
    SYNC1 = 9
    #: Paused in application
    PAUSED = 10
    #: Exited application
    FINISHED = 11
    #: Spare
    CPU_STATE_12 = 12
    #: Spare
    CPU_STATE_13 = 13
    #: Spare
    CPU_STATE_14 = 14
    #: Idle (SARK stub)
    IDLE = 15



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .eieio_message import AbstractEIEIOMessage
from .eieio_prefix import EIEIOPrefix
from .eieio_type import EIEIOType
from .create_eieio_command import read_eieio_command_message
from .create_eieio_data import read_eieio_data_message

__all__ = ["EIEIOPrefix", "EIEIOType", "read_eieio_command_message",
           "read_eieio_data_message", "AbstractEIEIOMessage"]



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.abstract_context_manager import AbstractContextManager


class Connection(AbstractContextManager, metaclass=AbstractBase):
    """
    An abstract connection to the SpiNNaker board over some medium.
    """

    __slots__ = ()

    @abstractmethod
    def is_connected(self) -> bool:
        """
        Determines if the medium is connected at this point in time.

        :return: True if the medium is connected, False otherwise
        :rtype: bool
        :raise SpinnmanIOException:
            If there is an error when determining the connectivity of the
            medium.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """
        Closes the connection.
        """
        raise NotImplementedError



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logging import getLogger
from typing import Generic, Optional, TypeVar, final
from spinn_utilities.abstract_base import abstractmethod
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from .scp_response import AbstractSCPResponse

logger = FormatAdapter(getLogger(__name__))
#: :meta private:
T = TypeVar("T")


class BMPResponse(AbstractSCPResponse, Generic[T]):
    """
    Represents an SCP response that's tailored for the BMP connection.
    """
    __slots__ = ("__operation", "__command", "__value")

    def __init__(self, operation: str, command: SCPCommand):
        super().__init__()
        self.__operation = operation
        self.__command = command
        self.__value: Optional[T] = None

    @property
    def _value(self) -> T:
        """
        The value parsed from the message. Subclasses have to expose this to
        make the parsed payload visible.
        """
        if self.__value is None:
            raise ValueError("no value parsed yet")
        return self.__value

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                self.__operation, self.__command.name, result.name)
        self.__value = self._parse_payload(data, offset)

    @abstractmethod
    def _parse_payload(self, data: bytes, offset: int) -> T:
        """
        Parse the payload of the message. The header will already be OK.

        :param data:
            The full content of the message
        :param offset:
            Where the payload should start in the message
        :return:
            The parsed payload
        """
        raise NotImplementedError


@final
class BMPOKResponse(BMPResponse[None]):
    """
    A BMP response without payload to parse.
    """
    @abstractmethod
    def _parse_payload(self, data: bytes, offset: int) -> None:
        if len(data) != offset:
            logger.warning("response message with unexpected extra {} bytes",
                           len(data) - offset)
        return None

    @property
    def _value(self) -> None:
        # Override to remove None check; this is always None
        return None



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import List
from spinn_utilities.overrides import overrides
from spinn_machine import RoutingEntry
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.messages.sdp import SDPHeader, SDPFlag

_ONE_WORD = struct.Struct("<I")


class _FixedRouteResponse(AbstractSCPResponse):
    """
    response for the fixed route read
    """
    __slots__ = [
        # the fixed route router entry
        "_route"]

    def __init__(self) -> None:
        super().__init__()
        self._route = 0

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Read Fixed Route entry", "CMD_RTR", result.name)

        self._route = _ONE_WORD.unpack_from(data, offset)[0]

    @property
    def route(self) -> RoutingEntry:
        """
        Converts this response into a Route

        :rtype: RoutingEntry
        """
        processor_ids: List[int] = list()
        for processor_id in range(26):
            if self._route & (1 << (6 + processor_id)) != 0:
                processor_ids.append(processor_id)
        link_ids: List[int] = list()
        for link_id in range(6):
            if self._route & (1 << link_id) != 0:
                link_ids.append(link_id)
        return RoutingEntry(processor_ids=processor_ids, link_ids=link_ids)


class FixedRouteRead(AbstractSCPRequest[_FixedRouteResponse]):
    """
    Gets a fixed route entry.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, app_id: int):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int y: The y-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int app_id:
            The ID of the application with which to associate the routes.
            If not specified, defaults to 0.
        :raise SpinnmanInvalidParameterException:
            * If x is out of range
            * If y is out of range
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR),
            argument_1=(app_id << 8) | 3, argument_2=1 << 31)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> _FixedRouteResponse:
        return _FixedRouteResponse()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import List, Union

from spinn_utilities.overrides import overrides
from spinnman.exceptions import (
    SpinnmanInvalidPacketException, SpinnmanInvalidParameterTypeException)
from spinnman.constants import EIEIO_COMMAND_IDS

from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader

_PATTERN_BB = struct.Struct("<BB")
_PATTERN_xxBBI = struct.Struct("<xxBBI")


class HostDataRead(EIEIOCommandMessage):
    """
    Packet sent by the host computer to the SpiNNaker system in the
    context of the buffering output technique to signal that the host has
    completed reading data from the output buffer, and that such space can
    be considered free to use again.
    """
    __slots__ = (
        "_acks",
        "_header")

    def __init__(
            self, n_requests: int, sequence_no: int,
            channel: Union[List[int], int], region_id: Union[List[int], int],
            space_read: Union[List[int], int]):
        """

        :param int n_requests:
        :param int sequence_no:
        :param channel:
        :type channel: list(int) or int
        :param region_id:
        :type region_id: list(int) or int
        :param space_read:
        :type space_read: list(int) or int
        """
        # pylint: disable=too-many-arguments
        if not isinstance(channel, list):
            channel = [channel]

        if not isinstance(region_id, list):
            region_id = [region_id]

        if not isinstance(space_read, list):
            space_read = [space_read]

        if len(channel) != n_requests or \
                len(region_id) != n_requests or \
                len(space_read) != n_requests:
            raise SpinnmanInvalidPacketException(
                "SpinnakerRequestReadData",
                "The format for a SpinnakerRequestReadData packet is "
                f"invalid: {n_requests} request(s), {len(space_read)} "
                f"space(s) read defined, {len(region_id)} region(s) defined, "
                f"{len(channel)} channel(s) defined")
        super().__init__(EIEIOCommandHeader(EIEIO_COMMAND_IDS.HOST_DATA_READ))
        self._header = _HostDataReadHeader(n_requests, sequence_no)
        self._acks = _HostDataReadAck(channel, region_id, space_read)

    @property
    def n_requests(self) -> int:
        """
        Gets the n_requests passed into the init.

        :rtype: int
        """
        return self._header.n_requests

    @property
    def sequence_no(self) -> int:
        """
        Gets the sequence_no passed into the init.

        :rtype: int
        """
        return self._header.sequence_no

    def channel(self, ack_id: int) -> int:
        """
        Gets the channel value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        return self._acks.channel(ack_id)

    def region_id(self, ack_id: int) -> int:
        """
        Gets the region_id value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        return self._acks.region_id(ack_id)

    def space_read(self, ack_id: int) -> int:
        """
        Gets the space_read value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        return self._acks.space_read(ack_id)

    @staticmethod
    @overrides(EIEIOCommandMessage.get_min_packet_length)
    def get_min_packet_length() -> int:
        return 8

    @staticmethod
    @overrides(EIEIOCommandMessage.from_bytestring)
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "HostDataRead":
        n_requests, sequence_no = _PATTERN_BB.unpack_from(data, offset)

        offset += 2
        n_requests &= 0x7
        channel = list()
        region_id = list()
        space_read = list()

        for _ in range(n_requests):
            channel_ack, region_id_ack, space_read_ack = \
                _PATTERN_xxBBI.unpack_from(data, offset)
            channel.append(channel_ack)
            region_id.append(region_id_ack)
            space_read.append(space_read_ack)
            offset += 8
        return HostDataRead(
            n_requests, sequence_no, channel, region_id, space_read)

    @property
    @overrides(EIEIOCommandMessage.bytestring)
    def bytestring(self) -> bytes:
        byte_string = super().bytestring
        n_requests = self.n_requests
        byte_string += _PATTERN_BB.pack(n_requests, self.sequence_no)
        for i in range(n_requests):
            byte_string += _PATTERN_xxBBI.pack(
                self.channel(i), self.region_id(i), self.space_read(i))
        return byte_string


class _HostDataReadHeader(object):
    """
    The HostDataRead contains itself on header with the number of requests
    and a sequence number.
    """
    __slots__ = [
        "_n_requests",
        "_sequence_no"]

    def __init__(self, n_requests: int, sequence_no: int):
        """

        :param int n_requests:
        :param int sequence_no:
        """
        self._n_requests = n_requests
        self._sequence_no = sequence_no

    @property
    def sequence_no(self) -> int:
        """
        Gets the sequence_no passed into the init.

        :rtype: int
        """
        return self._sequence_no

    @property
    def n_requests(self) -> int:
        """
        Gets the n_requests passed into the init.

        :rtype: int
        """
        return self._n_requests


class _HostDataReadAck(object):
    """
    Contains a set of ACKs which refer to each of the channels read.
    """
    __slots__ = [
        "_channel",
        "_region_id",
        "_space_read"]

    def __init__(self, channel: Union[List[int], int],
                 region_id: Union[List[int], int],
                 space_read: Union[List[int], int]):
        """

        :param channel:
        :type channel: list(int) or int
        :param region_id:
        :type region_id: list(int) or int
        :param space_read:
        :type space_read: list(int) or int
        """
        if not isinstance(channel, list):
            self._channel = [channel]
        else:
            self._channel = channel

        if not isinstance(region_id, list):
            self._region_id = [region_id]
        else:
            self._region_id = region_id

        if not isinstance(space_read, list):
            self._space_read = [space_read]
        else:
            self._space_read = space_read

    def channel(self, ack_id: int) -> int:
        """
        Gets the channel value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        if len(self._channel) > ack_id:
            return self._channel[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"channel ack_id needs to be comprised between 0 and "
            f"{len(self._channel) - 1:d}; current value: {ack_id:d}")

    def region_id(self, ack_id: int) -> int:
        """
        Gets the region_id value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        if len(self._region_id) > ack_id:
            return self._region_id[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"region ID ack_id needs to be comprised between 0 and "
            f"{len(self._region_id) - 1:d}; current value: {ack_id:d}")

    def space_read(self, ack_id: int) -> int:
        """
        Gets the space_read value for this ack_id.

        :param int ack_id:
        :rtype: int
        :raises SpinnmanInvalidParameterTypeException:
            If the ack_id is invalid
        """
        if len(self._space_read) > ack_id:
            return self._space_read[ack_id]
        raise SpinnmanInvalidParameterTypeException(
            "request_id", "integer",
            f"start address ack_id needs to be comprised between 0 and "
            f"{len(self._space_read) - 1:d}; current value: {ack_id:d}")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse


class RouterInit(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to initialise the router on a chip.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, n_entries: int, table_address: int,
                 base_address: int, app_id: int):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255
        :param int y: The y-coordinate of the chip, between 0 and 255
        :param int n_entries: The number of entries in the table, more than 0
        :param int table_address: The allocated table address
        :param int base_address: The base_address containing the entries
        :param int app_id:
            The ID of the application with which to associate the
            routes.  If not specified, defaults to 0.
        :raise SpinnmanInvalidParameterException:
            * If x is out of range
            * If y is out of range
            * If n_entries is 0 or less
            * If table_address is not positive
            * If base_address is not positive
        """
        # pylint: disable=too-many-arguments
        if n_entries < 1:
            raise SpinnmanInvalidParameterException(
                "n_entries", str(n_entries), "Must be more than 0")
        if base_address < 0:
            raise SpinnmanInvalidParameterException(
                "base_address", str(base_address),
                "Must be a positive integer")
        if table_address < 0:
            raise SpinnmanInvalidParameterException(
                "table_address", str(table_address),
                "Must be a positive integer")

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR),
            argument_1=((n_entries << 16) | (app_id << 8) | 2),
            argument_2=table_address, argument_3=base_address)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("RouterInit", "CMD_RTR")



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under
# return 0the License.

from typing import Dict, Iterable, Optional, Union
from spinn_utilities.overrides import overrides
from spinn_machine import CoreSubsets, Machine, virtual_machine
from spinnman.constants import N_RETRIES
from spinnman.exceptions import SpinnmanIOException
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.spinnaker_boot import SystemVariableDefinition
from spinnman.model import CPUInfos, VersionInfo
from spinnman.model.enums import CPUState
from spinnman.processes import ConnectionSelector
from .version5transceiver import Version5Transceiver


class Virtual5Transceiver(Version5Transceiver):
    """
    Overwrites the standard Version 5 Transceiver
    to intercept the sending of messages.

    This is just enough different to pass know tests.
    Extend for new tests!
    """

    @overrides(Version5Transceiver._boot_board)
    def _boot_board(self, extra_boot_values: Optional[Dict[
            SystemVariableDefinition, object]] = None):
        try:
            super()._boot_board(extra_boot_values)
        except SpinnmanIOException:
            pass

    @overrides(Version5Transceiver.read_memory)
    def read_memory(
            self, x: int, y: int, base_address: int, length: int,
            cpu: int = 0) -> bytearray:
        try:
            return super().read_memory(x, y, base_address, length, cpu)
        except SpinnmanIOException as exc:
            if (x == y == 255 and base_address == 4110450434):
                return bytearray(b'\x08\x08')
            raise NotImplementedError(
                f"Unexpected {x=} {y=} {base_address=}, {length=} {cpu=}") \
                from exc

    @overrides(Version5Transceiver._get_scamp_version)
    def _get_scamp_version(
            self, chip_x: int = AbstractSCPRequest.DEFAULT_DEST_X_COORD,
            chip_y: int = AbstractSCPRequest.DEFAULT_DEST_Y_COORD,
            connection_selector: Optional[ConnectionSelector] = None,
            n_retries: int = N_RETRIES) -> VersionInfo:
        try:
            return super()._get_scamp_version(
                chip_x, chip_y, connection_selector, n_retries)
        except SpinnmanIOException:
            version = VersionInfo(
                b'@\x00\x07\x08\xff\x00\x00\x00\x00\x00\x80\x00\x02\x00\x00\n'
                b'\x00\x00\x00\x01\xff\xffoa\xa3bSC&MP/SpiNNaker\x003.4.2\x00',
                offset=14)
            return version

    @overrides(Version5Transceiver.get_machine_details)
    def get_machine_details(self) -> Machine:
        try:
            return super().get_machine_details()
        except SpinnmanIOException:
            return virtual_machine(8, 8)

    @overrides(Version5Transceiver.get_cpu_infos)
    def get_cpu_infos(
            self, core_subsets: Optional[CoreSubsets] = None,
            states: Union[CPUState, Iterable[CPUState], None] = None,
            include: bool = True) -> CPUInfos:
        try:
            return super().get_cpu_infos(core_subsets, states, include)
        except SpinnmanIOException:
            # return an empty
            return CPUInfos()



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import traceback
from types import TracebackType
from typing import List, Optional, FrozenSet, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from spinnman.messages.scp.enums import SCPResult
    from spinnman.model.enums import CPUState
    from spinnman.model import CPUInfos
    from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
    from spinnman.connections.udp_packet_connections import SCAMPConnection


class SpinnmanException(Exception):
    """
    Superclass of exceptions that occur when dealing with communication
    with SpiNNaker.
    """


class SpinnmanInvalidPacketException(SpinnmanException):
    """
    An exception that indicates that a packet was not in the expected format.
    """

    def __init__(self, packet_type: str, problem: str):
        """
        :param str packet_type: The type of packet expected
        :param str problem: The problem with the packet
        """
        super().__init__(
            f"Invalid packet of type {packet_type} received: {problem}")
        self._packet_type = packet_type
        self._problem = problem

    @property
    def packet_type(self) -> str:
        """
        The packet type.

        :rtype: str
        """
        return self._packet_type

    @property
    def problem(self) -> str:
        """
        The problem with the packet.

        :rtype: str
        """
        return self._problem


class SpinnmanInvalidParameterException(SpinnmanException):
    """
    An exception that indicates that the value of one of the parameters
    passed was invalid.
    """

    def __init__(self, parameter: str, value, problem: str):
        """
        :param str parameter: The name of the parameter that is invalid
        :param str value: The value of the parameter that is invalid
        :param str problem: The problem with the parameter
        """
        super().__init__(
            f"Setting parameter {parameter} to value {value} is invalid: "
            f"{problem}")
        self._parameter = parameter
        self._value = value
        self._problem = problem

    @property
    def parameter(self) -> str:
        """
        The parameter with an invalid value.

        :rtype: str
        """
        return self._parameter

    @property
    def value(self):
        """
        The value that is invalid.
        """
        return self._value

    @property
    def problem(self) -> str:
        """
        The problem with the parameter value.

        :rtype: str
        """
        return self._problem


class SpinnmanInvalidParameterTypeException(SpinnmanException):
    """
    An exception that indicates that the type of one of the parameters
    passed was invalid.
    """

    def __init__(self, parameter: str, param_type, problem: str):
        """
        :param str parameter: The name of the parameter that is invalid
        :param str param_type: The type of the parameter that is invalid
        :param str problem: The problem with the parameter
        """
        super().__init__(
            f"Parameter {parameter} of type {param_type} is invalid: "
            f"{problem}")
        self._parameter = parameter
        self._type = param_type
        self._problem = problem

    @property
    def parameter(self) -> str:
        """
        The parameter with an invalid value.

        :rtype: str
        """
        return self._parameter

    @property
    def type(self):
        """
        The value that is invalid.
        """
        return self._type

    @property
    def problem(self) -> str:
        """
        The problem with the parameter value.

        :rtype: str
        """
        return self._problem


class SpinnmanIOException(SpinnmanException):
    """
    An exception that something went wrong with the underlying IO.
    """

    def __init__(self, problem: str):
        """
        :param str problem: The problem with the IO
        """
        super().__init__(f"IO Error: {problem}")
        self._problem = problem

    @property
    def problem(self) -> str:
        """
        The problem with IO.

        :rtype: str
        """
        return self._problem


class SpinnmanEOFException(SpinnmanIOException):
    """
    An exception that we're trying to do I/O on a closed socket.
    That isn't going to work!
    """

    def __init__(self) -> None:
        super().__init__("connection is closed")


class SpinnmanTimeoutException(SpinnmanException):
    """
    An exception that indicates that a timeout occurred before an operation
    could finish.
    """

    def __init__(self, operation: str, timeout: Optional[float],
                 msg: Optional[str] = None):
        """
        :param str operation: The operation being performed
        :param float timeout: The timeout value in seconds
        """
        if msg is None:
            msg = f"Operation {operation} timed out after {timeout} seconds"
        super().__init__(msg)

        self._operation = operation
        self._timeout = timeout

    @property
    def operation(self) -> str:
        """
        The operation that was performed.

        :rtype: str
        """
        return self._operation

    @property
    def timeout(self) -> Optional[float]:
        """
        The timeout value in seconds.

        :rtype: float
        """
        return self._timeout


class SpinnmanUnexpectedResponseCodeException(SpinnmanException):
    """
    Indicate that a response code returned from the board was unexpected
    for the current operation.
    """

    def __init__(self, operation: str, command,
                 response: Union[str, SCPResult]):
        """
        :param str operation: The operation being performed
        :param str command: The command being executed
        :param str response: The response received in error
        """
        super().__init__(
            f"Unexpected response {response} while performing "
            f"operation {operation} using command {command}")
        self._operation = operation
        self._command = command
        self._response = str(response)

    @property
    def operation(self) -> str:
        """
        The operation being performed.

        :rtype: str
        """
        return self._operation

    @property
    def command(self):
        """
        The command being executed.
        """
        return self._command

    @property
    def response(self) -> str:
        """
        The unexpected response.

        :rtype: str
        """
        return self._response


class SpinnmanGroupedProcessException(SpinnmanException):
    """
    Encapsulates exceptions from processes which communicate with a
    collection of cores/chips.
    """
    def __init__(self, error_requests: List[AbstractSCPRequest],
                 exceptions: List[Exception], tracebacks: List[TracebackType],
                 connections: List[SCAMPConnection]):
        problem = "Exceptions found were:\n"
        for error_request, exception, trace_back, connection in zip(
                error_requests, exceptions, tracebacks, connections):
            sdp_header = error_request.sdp_header
            phys_p = sdp_header.get_physical_cpu_id()
            location = f"board {connection.remote_ip_address} with ethernet " \
                       f"chip {connection.chip_x}:{connection.chip_y} " \
                       f"[{sdp_header.destination_chip_x}:" \
                       f"{sdp_header.destination_chip_y}:" \
                       f"{sdp_header.destination_cpu}({phys_p})]"
            problem += \
                f"   Received exception class: " \
                f"{exception.__class__.__name__}\n" \
                f"       With message {str(exception)}\n" \
                f"       When sending to {location}\n" \
                f"       Stack trace: {traceback.format_tb(trace_back)}\n"
        super().__init__(problem)


class SpinnmanGenericProcessException(SpinnmanException):
    """
    Encapsulates exceptions from processes which communicate with some
    core/chip.
    """
    def __init__(
            self, exception: Exception, tb: TracebackType,
            x: int, y: int, p: int, phys_p: str,
            tb2: Optional[TracebackType] = None):
        """
        :param Exception exception:
        :param int x:
        :param int y:
        :param int p:
        :param str phys_p:
        """
        # pylint: disable=too-many-arguments
        super().__init__(
            f"   Received exception class: {exception.__class__.__name__} \n"
            f"      With message: {str(exception)} \n"
            f"      When sending to {x}:{y}:{p}{phys_p}\n"
            f"      Stack trace: {traceback.format_tb(tb)}\n")

        self._stored_exception = exception
        if tb2 is not None:
            self.__traceback__ = tb2

    @property
    def exception(self) -> Exception:
        """
        :rtype: Exception
        """
        return self._stored_exception


class SpinnmanUnsupportedOperationException(SpinnmanException):
    """
    An exception that indicates that the given operation is not supported.
    """

    def __init__(self, operation: str):
        """
        :param str operation: The operation being requested
        """
        super().__init__(f"Operation {operation} is not supported")
        self._operation = operation

    @property
    def operation(self) -> str:
        """
        The unsupported operation requested.

        :rtype: str
        """
        return self._operation


class SpinnmanEIEIOPacketParsingException(SpinnmanException):
    """
    Unable to complete the parsing of the EIEIO packet received.
    The routine used is invalid or the content of the packet is invalid.
    """

    def __init__(self, parsing_format: str, packet: bytes):
        """
        :param str parsing_format:
        :param bytes packet:
        """
        super().__init__(
            "The packet received is being parsed as an EIEIO "
            f"{parsing_format} packet, "
            "but the content of the packet is invalid")
        self._packet = packet

    @property
    def packet(self) -> bytes:
        """
        :rtype: bytes
        """
        return self._packet


class SpiNNManCoresNotInStateException(SpinnmanTimeoutException):
    """
    Cores failed to reach a given state within a timeout.
    """

    def __init__(
            self, timeout: Optional[float],
            expected_states: FrozenSet[CPUState],
            failed_core_states: CPUInfos):
        """
        :param float timeout:
        :param set(CPUState) expected_states:
        :param CPUInfos failed_core_states:
        """
        n_cores = len(failed_core_states)
        if n_cores > 10:
            msg = (f"waiting for {n_cores} cores to reach "
                   f"one of {expected_states}")
        else:
            msg = (f"waiting for cores {failed_core_states} to reach "
                   f"one of {expected_states}")
        super().__init__(msg, timeout, msg)
        self._failed_core_states = failed_core_states

    def failed_core_states(self) -> CPUInfos:
        """
        :rtype: CPUInfos
        """
        return self._failed_core_states


class SpallocException(SpinnmanException):
    """
    Raised when there is a problem with the Spalloc session or job.
    """



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class RunTimeError(Enum):
    """
    SARK Run time errors.
    """
    #: No error
    NONE = 0
    #: Branch through zero
    RESET = 1
    #: Undefined instruction
    UNDEF = 2
    #: Undefined Supervisor Call instruction or no handler
    SVC = 3
    #: Prefetch abort
    PABT = 4
    #: Data abort
    DABT = 5
    #: Unhandled interrupt request
    IRQ = 6
    #: Unhandled Fast interrupt request
    FIQ = 7
    #: Unconfigured Vectored Interrupt Controller
    VIC = 8
    #: Generic user abort
    ABORT = 9
    #: "malloc" failure
    MALLOC = 10
    #: Divide by zero
    DIVBY0 = 11
    #: Event startup failure
    EVENT = 12
    #: Fatal Software error
    SWERR = 13
    #: Failed to allocate IO buffer
    IOBUF = 14
    #: Bad event enable
    ENABLE = 15
    #: Generic null pointer error
    NULL = 16
    #: Packet startup failure
    PKT = 17
    #: Timer startup failure
    TIMER = 18
    #: API startup failure
    API = 19
    #: Sark software version conflict
    SARK_VERSRION_INCORRECT = 20
    #: Unhandled exception
    UNRECOGNISED = 99999



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Callable, Optional

from spinn_utilities.overrides import overrides
from spinnman.messages.sdp import SDPMessage, SDPFlag
from spinnman.connections.abstract_classes import Listenable
from spinnman.exceptions import SpinnmanUnsupportedOperationException

from .udp_connection import UDPConnection

_TWO_SKIP = struct.Struct("<2x")


class SDPConnection(UDPConnection, Listenable[SDPMessage]):
    """
    A connection that talks SpiNNaker Datagram Protocol.
    """
    __slots__ = (
        "_chip_x",
        "_chip_y")

    def __init__(
            self, chip_x: int, chip_y: int,
            local_host: Optional[str] = None,
            local_port: Optional[int] = None,
            remote_host: Optional[str] = None,
            remote_port: Optional[int] = None):
        """
        :param int chip_x: The optional x-coordinate of the chip at the remote
            end of the connection. If not specified, it will not be possible
            to send SDP messages that require a response with this connection.
        :param int chip_y: The optional y-coordinate of the chip at the remote
            end of the connection. If not specified, it will not be possible
            to send SDP messages that require a response with this connection.
        :param str local_host: The optional IP address or host name of the
            local interface to listen on
        :param int local_port: The optional local port to listen on
        :param str remote_host: The optional remote host name or IP address to
            send messages to. If not specified, sending will not be possible
            using this connection
        :param int remote_port: The optional remote port number to send
            messages to. If not specified, sending will not be possible using
            this connection
        """
        # pylint: disable=too-many-arguments
        super().__init__(local_host, local_port, remote_host, remote_port)
        self._chip_x = chip_x
        self._chip_y = chip_y

    def receive_sdp_message(
            self, timeout: Optional[float] = None) -> SDPMessage:
        """
        Receives an SDP message from this connection.  Blocks until the
        message has been received, or a timeout occurs.

        :param int timeout:
            The time in seconds to wait for the message to arrive; if not
            specified, will wait forever, or until the connection is closed.
        :return: The received SDP message
        :rtype: SDPMessage
        :raise SpinnmanIOException:
            If there is an error receiving the message
        :raise SpinnmanTimeoutException:
            If there is a timeout before a message is received
        :raise SpinnmanInvalidPacketException:
            If the received packet is not a valid SDP message
        :raise SpinnmanInvalidParameterException:
            If one of the fields of the SDP message is invalid
        """
        data = self.receive(timeout)
        return SDPMessage.from_bytestring(data, 2)

    def send_sdp_message(self, sdp_message: SDPMessage):
        """
        Sends an SDP message down this connection.

        :param SDPMessage sdp_message: The SDP message to be sent
        :raise SpinnmanIOException:
            If there is an error sending the message.
        """
        if self._chip_x is None or self._chip_y is None:
            raise SpinnmanUnsupportedOperationException(
                "send on receive-only connection")
        # If a reply is expected, the connection should
        if sdp_message.sdp_header.flags == SDPFlag.REPLY_EXPECTED:
            sdp_message.sdp_header.update_for_send(self._chip_x, self._chip_y)
        else:
            sdp_message.sdp_header.update_for_send(0, 0)
        self.send(_TWO_SKIP.pack() + sdp_message.bytestring)

    @overrides(Listenable.get_receive_method)
    def get_receive_method(  # type: ignore[override]
            self) -> Callable[[], SDPMessage]:
        return self.receive_sdp_message

    def __repr__(self) -> str:
        return (
            f"SDPConnection(chip_x={self._chip_x}, chip_y={self._chip_y}, "
            f"local_host={self.local_ip_address}, local_port={self.local_port}"
            f", remote_host={self.remote_ip_address}, "
            f"remote_port={self.remote_port})")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Callable, Optional

from spinn_utilities.overrides import overrides

from spinnman.connections.abstract_classes import Listenable
from spinnman.messages.eieio import (
    read_eieio_command_message, read_eieio_data_message)
from spinnman.messages.eieio import AbstractEIEIOMessage

from .udp_connection import UDPConnection

_ONE_SHORT = struct.Struct("<H")
_REPR_TEMPLATE = "EIEIOConnection(local_host={}, local_port={},"\
    "remote_host={}, remote_port={})"


class EIEIOConnection(UDPConnection, Listenable[AbstractEIEIOMessage]):
    """
    A UDP connection for sending and receiving raw EIEIO messages.
    """
    __slots__ = ()

    def receive_eieio_message(
            self, timeout: Optional[float] = None) -> AbstractEIEIOMessage:
        """
        Receives an EIEIO message from this connection.  Blocks until
        a message has been received, or a timeout occurs.

        :param int timeout:
            The time in seconds to wait for the message to arrive; if not
            specified, will wait forever, or until the connection is closed
        :return: an EIEIO message
        :rtype: AbstractEIEIOMessage
        :raise SpinnmanIOException:
            If there is an error receiving the message.
        :raise SpinnmanTimeoutException:
            If there is a timeout before a message is received.
        :raise SpinnmanInvalidPacketException:
            If the received packet is not a valid EIEIO message.
        :raise SpinnmanInvalidParameterException:
            If one of the fields of the EIEIO message is invalid.
        """
        data = self.receive(timeout)
        header = _ONE_SHORT.unpack_from(data)[0]
        if header & 0xC000 == 0x4000:
            return read_eieio_command_message(data, 0)
        return read_eieio_data_message(data, 0)

    def send_eieio_message(self, eieio_message: AbstractEIEIOMessage):
        """
        Sends an EIEIO message down this connection.

        :param AbstractEIEIOMessage eieio_message:
            The EIEIO message to be sent
        :raise SpinnmanIOException:
            If there is an error sending the message
        """
        self.send(eieio_message.bytestring)

    def send_eieio_message_to(
            self, eieio_message: AbstractEIEIOMessage,
            ip_address: str, port: int):
        """

        :param AbstractEIEIOMessage eieio_message:
        :param str ip_address:
        :param int port:
        """
        self.send_to(eieio_message.bytestring, (ip_address, port))

    @overrides(Listenable.get_receive_method)
    def get_receive_method(self) -> Callable[  # type: ignore[override]
            [], AbstractEIEIOMessage]:
        return self.receive_eieio_message

    def __repr__(self) -> str:
        return _REPR_TEMPLATE.format(
            self.local_ip_address, self.local_port,
            self.remote_ip_address, self.remote_port)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import SCPCommand, SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException


_IPTAG_GET = 2

_IPTAG_FORMAT = struct.Struct("<4s6s3HIH3B")


class IPTagGetResponse(AbstractSCPResponse):
    """
    An SCP response to a request for an IP tags.
    """
    __slots__ = [
        "_count",
        "_flags",
        "_ip_address",
        "_mac_address",
        "_port",
        "_rx_port",
        "_spin_chip_x",
        "_spin_chip_y",
        "_spin_cpu",
        "_spin_port",
        "_timeout"]

    def __init__(self) -> None:
        super().__init__()
        self._ip_address = b'\0\0\0\0'
        self._mac_address = b'\0\0\0\0\0\0'
        self._port = 0
        self._timeout = 0
        self._flags = 0
        self._count = 0
        self._rx_port = 0
        self._spin_chip_y = 0
        self._spin_chip_x = 0
        self._spin_cpu = 0
        self._spin_port = 0

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "Get IP Tag Info", "CMD_IPTAG", result.name)
        (ip_address, mac_address, self._port, self._timeout,
         self._flags, self._count, self._rx_port, self._spin_chip_y,
         self._spin_chip_x, processor_and_port) = _IPTAG_FORMAT.unpack_from(
            data, offset)
        self._ip_address = bytearray(ip_address)
        self._mac_address = bytearray(mac_address)
        self._spin_port = (processor_and_port >> 5) & 0x7
        self._spin_cpu = processor_and_port & 0x1F

    @property
    def ip_address(self) -> bytes:
        """
        The IP address of the tag, as a bytearray of 4 bytes.

        :rtype: bytearray
        """
        return self._ip_address

    @property
    def mac_address(self) -> bytes:
        """
        The MAC address of the tag, as a bytearray of 6 bytes.

        :rtype: bytearray
        """
        return self._mac_address

    @property
    def port(self) -> int:
        """
        The port of the tag.

        :rtype: int
        """
        return self._port

    @property
    def timeout(self) -> int:
        """
        The timeout of the tag.

        :rtype: int
        """
        return self._timeout

    @property
    def flags(self) -> int:
        """
        The flags of the tag.

        :rtype: int
        """
        return self._flags

    @property
    def in_use(self) -> bool:
        """
        Whether the tag is marked as being in use.

        :rtype: bool
        """
        return (self._flags & 0x8000) > 0

    @property
    def is_temporary(self) -> bool:
        """
        Whether the tag is temporary.

        :rtype: bool
        """
        return (self._flags & 0x4000) > 0

    @property
    def is_arp(self) -> bool:
        """
        Whether the tag is in the Address Resolution Protocol state
        (where the MAC address is being looked up).

        .. note::
            This is a transient state; it is unlikely to be observed.

        :rtype: bool
        """
        return (self._flags & 0x2000) > 0

    @property
    def is_reverse(self) -> bool:
        """
        Whether the tag is a reverse tag.

        :rtype: bool
        """
        return (self._flags & 0x0200) > 0

    @property
    def strip_sdp(self) -> bool:
        """
        Whether the tag is to strip the SDP header.

        :rtype: bool
        """
        return (self._flags & 0x0100) > 0

    @property
    def count(self) -> int:
        """
        The count of the number of packets that have been sent with the tag.

        :rtype: int
        """
        return self._count

    @property
    def rx_port(self) -> int:
        """
        The receive port of the tag.

        :rtype: int
        """
        return self._rx_port

    @property
    def spin_chip_x(self) -> int:
        """
        The X-coordinate of the chip on which the tag is defined.

        :rtype: int
        """
        return self._spin_chip_x

    @property
    def spin_chip_y(self) -> int:
        """
        The Y-coordinate of the chip on which the tag is defined.

        :rtype: int
        """
        return self._spin_chip_y

    @property
    def spin_port(self) -> int:
        """
        The spin-port of the IP tag.

        :rtype: int
        """
        return self._spin_port

    @property
    def spin_cpu(self) -> int:
        """
        The CPU ID of the IP tag.

        :rtype: int
        """
        return self._spin_cpu


class IPTagGet(AbstractSCPRequest[IPTagGetResponse]):
    """
    An SCP Request to get an IP tag.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, tag: int):
        """
        :param int x: The x-coordinate of a chip, between 0 and 255
        :param int y: The y-coordinate of a chip, between 0 and 255
        :param int tag: The tag to get details of, between 0 and 7
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(_IPTAG_GET << 16) | tag,
            argument_2=1)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> IPTagGetResponse:
        return IPTagGetResponse()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast, Union
import struct
from enum import Enum
from spinnman.exceptions import SpinnmanInvalidParameterException

_ONE_SHORT = struct.Struct("<H")


class EIEIOCommandHeader(object):
    """
    EIEIO header for command packets.
    """
    __slots__ = "_command",

    def __init__(self, command: Union[int, Enum]):
        """

        :param command:
        :type command: int or Enum
        """
        if isinstance(command, Enum):
            command_value = command.value
        else:
            command_value = cast(int, command)
        if command_value < 0 or command_value >= 16384:
            raise SpinnmanInvalidParameterException(
                "command", command,
                "parameter command is outside the allowed range (0 to 16383)")
        self._command = command_value

    @property
    def command(self) -> int:
        """
        The command/ value of the command passed into the init.

        :rtype: int
        """
        return self._command

    @staticmethod
    def from_bytestring(data: bytes, offset: int) -> "EIEIOCommandHeader":
        """
        Read an EIEIO command header from a byte-string.

        :param data: The byte-string to read the data from
        :type data: bytes or bytearray
        :param int offset: The offset where the valid data starts
        :return: an EIEIO command header
        :rtype: EIEIOCommandHeader
        :raise SpinnmanIOException:
            If there is an error reading from the reader
        :raise SpinnmanInvalidParameterException:
            If there is an error setting any of the values
        """
        command_header = _ONE_SHORT.unpack_from(data, offset)[0]
        command = command_header & 0x3FFF

        return EIEIOCommandHeader(command)

    @property
    def bytestring(self) -> bytes:
        """
        The byte-string of the header.

        :rtype: bytes
        """
        return _ONE_SHORT.pack(0 << 15 | 1 << 14 | self._command)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional
from spinnman import constants

_PATTERN = struct.Struct(
    "<"   # Little-endian
    "8H"  # uint16_t adc[8]
    "4h"  # int16_t t_int[4]
    "4h"  # int16_t t_ext[4]
    "4h"  # int16_t fan[4]
    "I"   # uint32_t warning
    "I")  # uint32_t shutdown


class ADCInfo(object):
    """
    Container for the ADC data that's been retrieved from an FPGA.
    """
    __slots__ = [
        "_fan_0",
        "_fan_1",
        "_temp_btm",
        "_temp_ext_0",
        "_temp_ext_1",
        "_temp_top",
        "_voltage_1_2a",
        "_voltage_1_2b",
        "_voltage_1_2c",
        "_voltage_1_8",
        "_voltage_3_3",
        "_voltage_supply"]

    def __init__(self, adc_data: bytes, offset: int):
        """
        :param bytes adc_data:
            bytes from an SCP packet containing ADC information
        :raise SpinnmanInvalidParameterException:
            If the message does not contain valid ADC information
        """
        data = _PATTERN.unpack_from(adc_data, offset)

        self._voltage_1_2c = data[1] * constants.BMP_V_SCALE_2_5
        self._voltage_1_2b = data[2] * constants.BMP_V_SCALE_2_5
        self._voltage_1_2a = data[3] * constants.BMP_V_SCALE_2_5
        self._voltage_1_8 = data[4] * constants.BMP_V_SCALE_2_5
        self._voltage_3_3 = data[6] * constants.BMP_V_SCALE_3_3
        self._voltage_supply = data[7] * constants.BMP_V_SCALE_12
        self._temp_top = float(data[8]) * constants.BMP_TEMP_SCALE
        self._temp_btm = float(data[9]) * constants.BMP_TEMP_SCALE
        if data[12] != constants.BMP_MISSING_TEMP:
            self._temp_ext_0: Optional[float] = (
                float(data[12]) * constants.BMP_TEMP_SCALE)
        else:
            self._temp_ext_0 = None
        if data[13] != constants.BMP_MISSING_TEMP:
            self._temp_ext_1: Optional[float] = (
                float(data[13]) * constants.BMP_TEMP_SCALE)
        else:
            self._temp_ext_1 = None
        if data[16] != constants.BMP_MISSING_FAN:
            self._fan_0: Optional[float] = float(data[16])
        else:
            self._fan_0 = None
        if data[17] != constants.BMP_MISSING_FAN:
            self._fan_1: Optional[float] = float(data[17])
        else:
            self._fan_1 = None

    @property
    def voltage_1_2c(self) -> float:
        """
        Actual voltage of the 1.2V c supply rail.

        :rtype: float
        """
        return self._voltage_1_2c

    @property
    def voltage_1_2b(self) -> float:
        """
        Actual voltage of the 1.2V b supply rail.

        :rtype: float
        """
        return self._voltage_1_2b

    @property
    def voltage_1_2a(self) -> float:
        """
        Actual voltage of the 1.2V a supply rail.

        :rtype: float
        """
        return self._voltage_1_2a

    @property
    def voltage_1_8(self) -> float:
        """
        Actual voltage of the 1.8V supply rail.

        :rtype: float
        """
        return self._voltage_1_8

    @property
    def voltage_3_3(self) -> float:
        """
        Actual voltage of the 3.3V supply rail.

        :rtype: float
        """
        return self._voltage_3_3

    @property
    def voltage_supply(self) -> float:
        """
        Actual voltage of the main power supply (nominally 12V).

        :rtype: float
        """
        return self._voltage_supply

    @property
    def temp_top(self) -> float:
        """
        Temperature top.

        :rtype: float
        """
        return self._temp_top

    @property
    def temp_btm(self) -> float:
        """
        Temperature bottom.

        :rtype: float
        """
        return self._temp_btm

    @property
    def temp_ext_0(self) -> Optional[float]:
        """
        Temperature external 0.

        :rtype: float or None
        """
        return self._temp_ext_0

    @property
    def temp_ext_1(self) -> Optional[float]:
        """
        Temperature external 1.

        :rtype: float or None
        """
        return self._temp_ext_1

    @property
    def fan_0(self) -> Optional[float]:
        """
        Fan 0.

        :rtype: float or None
        """
        return self._fan_0

    @property
    def fan_1(self) -> Optional[float]:
        """
        Fan 1.

        :rtype: float or None
        """
        return self._fan_1



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from enum import Enum
from typing import List, Type, TypeVar
from spinnman.model.enums import (
    DiagnosticFilterDestination, DiagnosticFilterSource,
    DiagnosticFilterPayloadStatus, DiagnosticFilterDefaultRoutingStatus,
    DiagnosticFilterEmergencyRoutingStatus, DiagnosticFilterPacketType)
#: :meta private:
E = TypeVar("E", bound=Enum)

# Bit offsets of the various fields in the filter word
_PACKET_TYPE_OFFSET = 0
_EMERGENCY_ROUTE_OFFSET = 4
_EMERGENCY_ROUTE_MODE_OFFSET = 8
_DEFAULT_ROUTE_OFFSET = 10
_PAYLOAD_OFFSET = 12
_SOURCE_OFFSET = 14
_DESTINATION_OFFSET = 16
_ENABLE_INTERRUPT_OFFSET = 30


# Uses an enum to set flags in the filter word from a given offset
def _set_flags_in_word(
        word: int, enum_list: List[E], enum_type: Type[E],
        offset: int) -> int:
    if enum_list is None:
        enum_values: List[E] = list()
    else:
        enum_values = list(enum_list)
    if not enum_values:
        enum_values = [value for value in enum_type]
    for enum_value in enum_values:
        word |= 1 << (enum_value.value + offset)
    return word


# Uses an enum to read flags in the filter word from a given offset
def _read_flags_from_word(
        word: int, enum_list: Type[E], offset: int) -> List[E]:
    flags: List[E] = list()
    for enum_value in enum_list:
        if word & 1 << (enum_value.value + offset) != 0:
            flags.append(enum_value)
    return flags


class DiagnosticFilter(object):
    """
    A router diagnostic counter filter, which counts packets passing
    through the router with certain properties.  The counter will be
    incremented so long as the packet matches one of the values in each
    field i.e. one of each of the destinations, sources, payload_statuses,
    default_routing_statuses, emergency_routing_statuses and packet_types.
    """
    __slots__ = [
        "_default_routing_statuses",
        "_destinations",
        "_emergency_routing_statuses",
        "_enable_interrupt_on_counter_event",
        "_match_emergency_routing_status_to_incoming_packet",
        "_packet_types",
        "_payload_statuses",
        "_sources"]

    def __init__(self, enable_interrupt_on_counter_event: bool,
                 match_emergency_routing_status_to_incoming_packet: bool,
                 destinations: List[DiagnosticFilterDestination],
                 sources: List[DiagnosticFilterSource],
                 payload_statuses: List[DiagnosticFilterPayloadStatus],
                 default_routing_statuses: List[
                     DiagnosticFilterDefaultRoutingStatus],
                 emergency_routing_statuses: List[
                     DiagnosticFilterEmergencyRoutingStatus],
                 packet_types: List[DiagnosticFilterPacketType]):
        """
        :param bool enable_interrupt_on_counter_event: Indicates whether
            an interrupt should be raised when this rule matches
        :param bool match_emergency_routing_status_to_incoming_packet:
            Indicates whether the emergency routing statuses should be matched
            against packets arriving at this router (if True), or if they
            should be matched against packets leaving this router (if False)
        :param list(DiagnosticFilterDestination) destinations:
            Increment the counter if one or more of the given destinations
            match
        :param list(DiagnosticFilterSource) sources:
            Increment the counter if one or more of the given sources match
            (or `None` or empty list to match all)
        :param list(DiagnosticFilterPayloadStatus) payload_statuses:
            Increment the counter if one or more of the given payload statuses
            match (or `None` or empty list to match all)
        :param default_routing_statuses:
            Increment the counter if one or more of the given default routing
            statuses match (or `None` or empty list to match all)
        :type default_routing_statuses:
            list(DiagnosticFilterDefaultRoutingStatus)
        :param emergency_routing_statuses:
            Increment the counter if one or more of the given emergency routing
            statuses match (or `None` or empty list to match all)
        :type emergency_routing_statuses:
            list(DiagnosticFilterEmergencyRoutingStatus)
        :param list(DiagnosticFilterPacketType) packet_types:
            Increment the counter if one or more of the given packet types
            match (or `None` or empty list to match all)
        """
        # pylint: disable=too-many-arguments
        self._enable_interrupt_on_counter_event = \
            enable_interrupt_on_counter_event
        self._match_emergency_routing_status_to_incoming_packet = \
            match_emergency_routing_status_to_incoming_packet
        self._destinations = destinations
        self._sources = sources
        self._payload_statuses = payload_statuses
        self._default_routing_statuses = default_routing_statuses
        self._emergency_routing_statuses = emergency_routing_statuses
        self._packet_types = packet_types

    @property
    def enable_interrupt_on_counter_event(self) -> bool:
        """
        Returns the enable interrupt on counter event passed into
        the init unchanged

        Currently unused
        """
        return self._enable_interrupt_on_counter_event

    @property
    def match_emergency_routing_status_to_incoming_packet(self) -> bool:
        """
        Returns the match emergency routing status to incoming packet passed
        into the init unchanged

        Currently unused
        """
        return self._match_emergency_routing_status_to_incoming_packet

    @property
    def destinations(self) -> List[DiagnosticFilterDestination]:
        """
        Returns the destinations passed into the init unchanged

        Currently unused
        """
        return self._destinations

    @property
    def sources(self) -> List[DiagnosticFilterSource]:
        """
        Returns the sources passed into the init unchanged

        Currently unused
        """
        return self._sources

    @property
    def payload_statuses(self) -> List[DiagnosticFilterPayloadStatus]:
        """
        Returns the payload statuses passed into the init unchanged

        Currently unused
        """
        return self._payload_statuses

    @property
    def default_routing_statuses(self) -> List[
            DiagnosticFilterDefaultRoutingStatus]:
        """
        Returns the default routing statuses passed into the init unchanged

        Currently unused
        """
        return self._default_routing_statuses

    @property
    def emergency_routing_statuses(self) -> List[
            DiagnosticFilterEmergencyRoutingStatus]:
        """
        Returns the emergency routing statuses passed into the init unchanged

        Currently unused
        """
        return self._emergency_routing_statuses

    @property
    def packet_types(self) -> List[DiagnosticFilterPacketType]:
        """
        Returns the packet types passed into the init unchanged

        Currently unused
        """
        return self._packet_types

    @property
    def filter_word(self) -> int:
        """
        A word of data that can be written to the router to set up the filter.
        """
        data = 0
        if self._enable_interrupt_on_counter_event:
            data |= 1 << _ENABLE_INTERRUPT_OFFSET
        if not self._match_emergency_routing_status_to_incoming_packet:
            data |= 1 << _EMERGENCY_ROUTE_MODE_OFFSET
        data = _set_flags_in_word(data, self._destinations,
                                  DiagnosticFilterDestination,
                                  _DESTINATION_OFFSET)
        data = _set_flags_in_word(data, self._sources,
                                  DiagnosticFilterSource,
                                  _SOURCE_OFFSET)
        data = _set_flags_in_word(data, self._payload_statuses,
                                  DiagnosticFilterPayloadStatus,
                                  _PAYLOAD_OFFSET)
        data = _set_flags_in_word(data, self._default_routing_statuses,
                                  DiagnosticFilterDefaultRoutingStatus,
                                  _DEFAULT_ROUTE_OFFSET)
        data = _set_flags_in_word(data, self._emergency_routing_statuses,
                                  DiagnosticFilterEmergencyRoutingStatus,
                                  _EMERGENCY_ROUTE_OFFSET)
        data = _set_flags_in_word(data, self._packet_types,
                                  DiagnosticFilterPacketType,
                                  _PACKET_TYPE_OFFSET)

        return data

    @staticmethod
    def read_from_int(int_value: int) -> DiagnosticFilter:
        """
        Claims to returns a filter that reads an int

        Currently only called by unused Transceiver methods

        :param int int_value:
        :rtype: DiagnosticFilter
        """
        enable_interrupt_on_counter_event = (
            (int_value >> _ENABLE_INTERRUPT_OFFSET) & 0x1) == 1
        match_emergency_routing_status_to_incoming_packet = (
            (int_value >> _EMERGENCY_ROUTE_MODE_OFFSET) & 0x1) == 0
        destinations = _read_flags_from_word(
            int_value, DiagnosticFilterDestination, _DESTINATION_OFFSET)
        sources = _read_flags_from_word(
            int_value, DiagnosticFilterSource, _SOURCE_OFFSET)
        payload_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterPayloadStatus, _PAYLOAD_OFFSET)
        default_routing_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterDefaultRoutingStatus,
            _DEFAULT_ROUTE_OFFSET)
        emergency_routing_statuses = _read_flags_from_word(
            int_value, DiagnosticFilterEmergencyRoutingStatus,
            _EMERGENCY_ROUTE_OFFSET)
        packet_types = _read_flags_from_word(
            int_value, DiagnosticFilterPacketType, _PACKET_TYPE_OFFSET)

        return DiagnosticFilter(
            enable_interrupt_on_counter_event,
            match_emergency_routing_status_to_incoming_packet,
            destinations, sources, payload_statuses, default_routing_statuses,
            emergency_routing_statuses, packet_types)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from typing import cast
from spinn_machine import CoreSubsets
from spinnman.model import CPUInfos
from spinnman.model.cpu_info import CPUInfo, _vcpu_t, _VCPU_PATTERN
from spinnman.constants import CPU_INFO_BYTES
from spinnman.utilities.utility_functions import get_vcpu_address
from spinnman.messages.scp.impl.read_memory import ReadMemory, Response
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class GetCPUInfoProcess(AbstractMultiConnectionProcess[Response]):
    """
    Gets the CPU for processors over the provided connection.

    This base class returns info for all states.
    """
    __slots__ = ("__cpu_infos", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self.__cpu_infos = CPUInfos()

    def _is_desired(self, cpu_info: CPUInfo) -> bool:
        # pylint: disable=unused-argument
        return True

    def __handle_response(self, x: int, y: int, p: int, response: Response):
        cpu_data = cast(_vcpu_t, _VCPU_PATTERN.unpack_from(
            response.data, response.offset))
        cpu_info = CPUInfo(x, y, p, cpu_data)
        if self._is_desired(cpu_info):
            self.__cpu_infos.add_info(cpu_info)

    def get_cpu_info(self, core_subsets: CoreSubsets) -> CPUInfos:
        """
        :param ~spinn_machine.CoreSubsets core_subsets:
        :rtype: CPUInfos
        """
        with self._collect_responses():
            for core_subset in core_subsets:
                x, y = core_subset.x, core_subset.y
                for p in core_subset.processor_ids:
                    self._send_request(
                        ReadMemory((x, y, 0), get_vcpu_address(p),
                                   CPU_INFO_BYTES),
                        functools.partial(self.__handle_response, x, y, p))

        return self.__cpu_infos



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
API of the client for the Spalloc web service.
"""

import struct
from typing import Optional, Tuple
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XY
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.exceptions import SpinnmanTimeoutException
from spinnman.messages.eieio import (
    AbstractEIEIOMessage,
    read_eieio_command_message, read_eieio_data_message)
from spinnman.messages.sdp import SDPMessage, SDPFlag, SDPHeader
from spinnman.messages.scp.impl import IPTagSet
from .spalloc_proxied_connection import SpallocProxiedConnection
# mypy: disable-error-code=empty-body

_ONE_SHORT = struct.Struct("<H")
_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP: bytes = b'\0\0'
_NUM_UPDATE_TAG_TRIES = 3
_UPDATE_TAG_TIMEOUT = 1.0


class SpallocEIEIOListener(
        EIEIOConnection, SpallocProxiedConnection, metaclass=AbstractBase):
    """
    The socket interface supported by proxied EIEIO listener sockets.
    This emulates an :py:class:`EIEOConnection` opened with no address
    specified.
    """
    __slots__ = ()

    @overrides(EIEIOConnection.receive_eieio_message)
    def receive_eieio_message(
            self, timeout: Optional[float] = None) -> AbstractEIEIOMessage:
        data = self.receive(timeout)
        header = _ONE_SHORT.unpack_from(data)[0]
        if header & 0xC000 == 0x4000:
            return read_eieio_command_message(data, 0)
        return read_eieio_data_message(data, 0)

    @overrides(SpallocProxiedConnection.send)
    def send(self, data: bytes):
        """
        .. note::
            This class does not allow sending.
        """

    @abstractmethod
    def _get_chip_coords(self, ip_address: str) -> XY:
        """
        Get the coordinates of a chip given its IP address.

        :param str ip_address:
            The IP address of an Ethernet-enabled chip in the job.
        :return: Ethernet-enabled chip coordinates: X, Y
        :rtype: tuple(int, int)
        """
        raise NotImplementedError

    @abstractmethod
    def send_to_chip(
            self, message: bytes, x: int, y: int, port: int = SCP_SCAMP_PORT):
        """
        Send a message on an open socket to a particular board.

        :param bytes message: The message to send.
        :param int x:
            The X coordinate of the Ethernet-enabled chip to send the message
            to.
        :param int y:
            The Y coordinate of the Ethernet-enabled chip to send the message
            to.
        :param int port:
            The UDP port on the Ethernet-enabled chip to send the message to.
            Defaults to the SCP port.
        """
        raise NotImplementedError

    def send_to(self, data: bytes, address: Tuple[str, int]):
        """
        Send a message on an open socket.

        :param bytes message: The message to send.
        :param tuple(str,int) address:
            Where to send it to. Must be the address of an Ethernet-enabled
            chip on a board allocated to the job. Does not mean that SpiNNaker
            is listening on that port (but the SCP port is being listened to if
            the board is booted).
        """
        ip, port = address
        x, y = self._get_chip_coords(ip)
        self.send_to_chip(data, x, y, port)

    @property
    @abstractmethod
    def local_ip_address(self) -> str:  # type: ignore[override]
        """
        The IP address on the server to which the connection is bound.

        :return: The IP address as a dotted string, e.g., 0.0.0.0
        :rtype: str
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def local_port(self) -> int:  # type: ignore[override]
        """
        The port on the server to which the connection is bound.

        :return: The local port number
        :rtype: int
        """
        raise NotImplementedError

    def send_eieio_message_to_core(
            self, eieio_message: AbstractEIEIOMessage, x: int, y: int, p: int,
            ip_address: str):
        """
        Send an EIEIO message (one way) to a given core.

        :param AbstractEIEIOMessage eieio_message:
            The message to send.
        :param int x:
            The X coordinate of the core to send to.
        :param int y:
            The Y coordinate of the core to send to.
        :param int p:
            The ID of the core to send to.
        :param str ip_address:
            The IP address of the Ethernet-enabled chip to route the message
            via.
        """
        sdp_message = SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED, tag=0,
                destination_port=1, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y,
                source_port=0, source_cpu=0,
                source_chip_x=0, source_chip_y=0),
            data=eieio_message.bytestring)
        self.send_to(
            _TWO_SKIP + sdp_message.bytestring, (ip_address, SCP_SCAMP_PORT))

    def update_tag(self, x: int, y: int, tag: int, do_receive: bool = True):
        """
        Update the given tag on the given Ethernet-enabled chip to send
        messages to this connection.

        :param int x: The Ethernet-enabled chip's X coordinate
        :param int y: The Ethernet-enabled chip's Y coordinate
        :param int tag: The tag ID to update
        :param bool do_receive: Whether to receive the response or not
        :raises SpinnmanTimeoutException:
            If the message isn't handled within a reasonable timeout.
        :raises SpinnmanUnexpectedResponseCodeException:
            If the message is rejected by SpiNNaker/SCAMP.
        """
        request = IPTagSet(
            x, y, [0, 0, 0, 0], 0, tag, strip=True, use_sender=True)
        request.sdp_header.flags = SDPFlag.REPLY_EXPECTED_NO_P2P
        request.sdp_header.update_for_send(x, y)
        data = _TWO_SKIP + request.bytestring
        for _try in range(_NUM_UPDATE_TAG_TRIES):
            try:
                self.send_to_chip(data, x, y, SCP_SCAMP_PORT)
                if do_receive:
                    response_data = self.receive(_UPDATE_TAG_TIMEOUT)
                    request.get_scp_response().read_bytestring(
                        response_data, len(_TWO_SKIP))
                return
            except SpinnmanTimeoutException as e:
                if _try + 1 == _NUM_UPDATE_TAG_TRIES:
                    raise e

    def update_tag_by_ip(self, ip_address: str, tag: int):
        """
        Update a tag on a board at a given IP address to send messages to this
        connection.

        :param str ip_address: The address of the Ethernet-enabled chip
        :param int tag: The ID of the tag
        """
        x, y = self._get_chip_coords(ip_address)
        self.update_tag(x, y, tag)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class SCPCommand(Enum):
    """
    The SCP Commands.
    """
    # ======== SARK ========
    #: Get software version.
    CMD_VER = 0
    CMD_RUN = 1
    #: Read SDRAM.
    CMD_READ = 2
    #: Write SDRAM.
    CMD_WRITE = 3
    CMD_APLX = 4
    CMD_FILL = 5
    # ======== SCAMP ========
    CMD_COUNT = 15
    CMD_REMAP = 16
    #: Read neighbouring chip's memory.
    CMD_LINK_READ = 17
    #: Write neighbouring chip's memory.
    CMD_LINK_WRITE = 18
    CMD_AR = 19
    #: Send a Nearest-Neighbour packet.
    CMD_NNP = 20
    #: Copy a binary from an adjacent chip and start it
    CMD_APP_COPY_RUN = 21
    #: Send a Signal.
    CMD_SIG = 22
    #: Send Flood-Fill Data.
    CMD_FFD = 23
    CMD_AS = 24
    #: Control the LEDs.
    CMD_LED = 25
    #: Set an IPTAG.
    CMD_IPTAG = 26
    CMD_SROM = 27
    #: Allocate or Free SDRAM or Routing entries.
    CMD_ALLOC = 28
    #: Initialise the router.
    CMD_RTR = 29
    #: Dropped Packet Reinjection setup.
    CMD_DPRI = 30
    #: Get Chip Summary Information.
    CMD_INFO = 31
    #: Control sending of synchronisation messages.
    CMD_SYNC = 32

    # ======== BMP ========
    #: Get BMP info structures
    CMD_BMP_INFO = 48
    CMD_FLASH_COPY = 49
    CMD_FLASH_ERASE = 50
    CMD_FLASH_WRITE = 51
    CMD_RESET = 55
    #: Turns on or off the machine via BMP.
    CMD_BMP_POWER = 57

    # ======== OUTBOUND (OBSOLETE) ========
    CMD_TUBE = 64



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .eieio_command_header import EIEIOCommandHeader
from .eieio_command_message import EIEIOCommandMessage
from .event_stop_request import EventStopRequest
from .host_data_read import HostDataRead
from .host_data_read_ack import HostDataReadAck
from .host_send_sequenced_data import HostSendSequencedData
from .notification_protocol_db_location import (
    NotificationProtocolDatabaseLocation)
from .notification_protocol_pause_stop import NotificationProtocolPauseStop
from .notification_protocol_start_resume import NotificationProtocolStartResume
from .padding_request import PaddingRequest
from .spinnaker_request_buffers import SpinnakerRequestBuffers
from .spinnaker_request_read_data import SpinnakerRequestReadData
from .start_requests import StartRequests
from .stop_requests import StopRequests

__all__ = ["EIEIOCommandHeader", "EIEIOCommandMessage",
           "EventStopRequest", "HostDataRead", "HostSendSequencedData",
           "NotificationProtocolDatabaseLocation",
           "NotificationProtocolPauseStop", "NotificationProtocolStartResume",
           "PaddingRequest", "SpinnakerRequestBuffers", "HostDataReadAck",
           "SpinnakerRequestReadData", "StartRequests", "StopRequests"]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional, Sequence, Tuple

from spinn_utilities.overrides import overrides

from spinnman.constants import SCP_SCAMP_PORT
from spinnman.messages.scp.enums import SCPResult
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.connections.abstract_classes import AbstractSCPConnection
from spinnman.model import BMPConnectionData

from .udp_connection import UDPConnection

_TWO_SHORTS = struct.Struct("<2H")
_TWO_SKIP = struct.Struct("<2x")


class BMPConnection(UDPConnection, AbstractSCPConnection):
    """
    A BMP connection which supports queries to the BMP of a SpiNNaker machine.
    """
    __slots__ = ("_boards", )

    def __init__(self, connection_data: BMPConnectionData):
        """
        :param BMPConnectionData connection_data:
            The description of what to connect to.
        """
        port = SCP_SCAMP_PORT if connection_data.port_num is None\
            else connection_data.port_num
        super().__init__(
            remote_host=connection_data.ip_address, remote_port=port)
        self._boards = connection_data.boards

    @property
    def boards(self) -> Sequence[int]:
        """
        The set of boards supported by the BMP.

        :rtype: iterable(int)
        """
        return self._boards

    @property
    @overrides(AbstractSCPConnection.chip_x, extend_doc=False)
    def chip_x(self) -> int:
        """
        Defined to satisfy the AbstractSCPConnection - always 0 for a BMP.
        """
        return 0

    @property
    @overrides(AbstractSCPConnection.chip_y, extend_doc=False)
    def chip_y(self) -> int:
        """
        Defined to satisfy the AbstractSCPConnection - always 0 for a BMP.
        """
        return 0

    @overrides(AbstractSCPConnection.get_scp_data)
    def get_scp_data(self, scp_request: AbstractSCPRequest) -> bytes:
        scp_request.sdp_header.update_for_send(0, 0)
        return _TWO_SKIP.pack() + scp_request.bytestring

    @overrides(AbstractSCPConnection.receive_scp_response)
    def receive_scp_response(self, timeout: Optional[float] = 1.0) -> Tuple[
            SCPResult, int, bytes, int]:
        data = self.receive(timeout)
        result, sequence = _TWO_SHORTS.unpack_from(data, 10)
        return SCPResult(result), sequence, data, 2

    def __repr__(self):
        return (
            f"BMPConnection("
            f"boards={self._boards}, local_host={self.local_ip_address}, "
            f"local_port={self.local_port}, "
            f"remote_host={self.remote_ip_address}, "
            f"remote_port={self.remote_port})")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.enums import SCPResult
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException


class CheckOKResponse(AbstractSCPResponse):
    """
    An SCP response to a request which returns nothing other than OK.
    """
    __slots__ = (
        "_command",
        "_operation")

    def __init__(self, operation: str, command):
        """
        :param str operation: The operation being performed
        :param command: The command that was sent
        :type command: str or ~enum.Enum or int
        """
        super().__init__()
        self._operation = operation
        self._command = command

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                self._operation, str(self._command), result.name)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Union
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest)
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.scp.impl.check_ok_response import CheckOKResponse
from spinnman.messages.scp.enums.led_action import LEDAction


class BMPSetLed(BMPRequest):
    """
    Set the LED(s) of a board to either on, off or toggling.

    This class is currently deprecated and untested as there is no
    known use except for Transceiver.set_led which is itself deprecated.
    """
    __slots__ = ()

    def __init__(self, led: Union[int, Iterable[int]], action: LEDAction,
                 boards: Union[int, Iterable[int]]):
        """
        :param led: Number of the LED or an iterable of LEDs to set the
            state of (0-7)
        :type led: int or list(int)
        :param LEDAction action:
            State to set the LED to, either on, off or toggle
        :param boards: Specifies the board to control the LEDs of. This may
            also be an iterable of multiple boards (in the same frame).
        :type boards: int or list(int)
        """
        # set up the led entry for arg1
        if isinstance(led, int):
            arg1 = action.value << (led * 2)
        else:
            arg1 = sum(action.value << (a_led * 2) for a_led in led)

        # Bitmask of boards to control
        arg2 = self.get_board_mask(boards)

        # initialise the request now
        super().__init__(
            boards,
            SCPRequestHeader(command=SCPCommand.CMD_LED),
            argument_1=arg1, argument_2=arg2)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Set the LEDs of a board", "CMD_LED")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_machine import Router, RoutingEntry
from spinnman.messages.scp.impl import FixedRouteInit
from spinnman.processes import AbstractMultiConnectionProcess


class LoadFixedRouteRoutingEntryProcess(AbstractMultiConnectionProcess):
    """
    Load a fixed route routing entry onto a chip.
    """
    __slots__ = ()

    def load_fixed_route(
            self, x: int, y: int, fixed_route: RoutingEntry,
            app_id: int = 0):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions.
        :param int y: The y-coordinate of the chip, between 0 and 255;
            this is not checked due to speed restrictions.
        :param ~spinn_machine.RoutingEntry fixed_route:
            the fixed route entry
        :param int app_id: The ID of the application with which to associate
            the routes.  If not specified, defaults to 0.
        """
        route_entry = Router.convert_routing_table_entry_to_spinnaker_route(
            fixed_route)
        with self._collect_responses():
            self._send_request(FixedRouteInit(x, y, route_entry, app_id))



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional
from spinn_utilities.overrides import overrides
from spinnman.connections.abstract_classes import Connection
from spinnman.data import SpiNNManDataView
from spinnman.transceiver.base_transceiver import BaseTransceiver


class Version5Transceiver(BaseTransceiver):
    """
    Implementation of the Transceiver classes for Version 5 boards

    This class should ONLY be created via by transceiver_factory.py
    """

    @overrides(BaseTransceiver.__init__)
    def __init__(self, connections: Optional[List[Connection]] = None,
                 power_cycle: bool = False):
        super().__init__(connections, power_cycle)
        assert SpiNNManDataView.get_machine_version().number == 5

    @property
    @overrides(BaseTransceiver.boot_led_0_value)
    def boot_led_0_value(self) -> int:
        return 0x00000001



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest, BMPOKResponse)
from spinnman.messages.scp.abstract_messages.bmp_request import Boards
from spinnman.messages.scp.enums import SCPCommand, PowerCommand
from spinnman.messages.scp import SCPRequestHeader

logger = FormatAdapter(logging.getLogger(__name__))


class SetPower(BMPRequest[BMPOKResponse]):
    """
    An SCP request for the BMP to power on or power off a rack of boards.
    """
    __slots__ = ()

    def __init__(
            self, power_command: PowerCommand, boards: Boards, *,
            delay=0.0, board_to_send_to=0):
        """
        .. note::
            There is currently a bug in the BMP that means some boards don't
            respond to power commands not sent to BMP 0. Thus changing the
            board_to_send_to parameter is not recommended!

        :param PowerCommand power_command: The power command being sent
        :param boards: The boards on the same backplane to power on or off
        :type boards: int or list(int)
        :param float delay:
            Number of seconds delay between power state changes of
            the different boards.
        :param int board_to_send_to:
            The optional board to send the command to if this is to be sent
            to a frame of boards.

            .. note::
                Leave this at the default because of hardware bugs.
        """
        if board_to_send_to != 0:
            logger.warning(
                "There is currently a bug in the BMP that means some boards"
                " don't respond to power commands not sent to BMP 0.  Thus "
                "changing the board_to_send_to is not recommended!")

        arg1 = (int(delay * 1000) << 16) | power_command.value
        arg2 = self.get_board_mask(boards)

        super().__init__(
            board_to_send_to,
            SCPRequestHeader(command=SCPCommand.CMD_BMP_POWER),
            argument_1=arg1, argument_2=arg2)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> BMPOKResponse:
        return BMPOKResponse("powering request", SCPCommand.CMD_BMP_POWER)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Tuple
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.messages.scp.enums import SCPResult
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from .connection import Connection


class AbstractSCPConnection(Connection, metaclass=AbstractBase):
    """
    A sender and receiver of SCP messages.
    """

    __slots__ = ()

    @abstractmethod
    def is_ready_to_receive(self, timeout: float = 0):
        """
        Determines if there is an SCP packet to be read without blocking.

        :param int timeout:
            The time to wait before returning if the connection is not ready
        :return: True if there is an SCP packet to be read
        :rtype: bool
        """
        raise NotImplementedError

    @abstractmethod
    def receive_scp_response(self, timeout: Optional[float] = 1.0) -> Tuple[
            SCPResult, int, bytes, int]:
        """
        Receives an SCP response from this connection.  Blocks
        until a message has been received, or a timeout occurs.

        :param int timeout:
            The time in seconds to wait for the message to arrive; if `None`,
            will wait forever, or until the connection is closed
        :return: The SCP result, the sequence number, the data of the response
            and the offset at which the data starts (i.e., where the SDP
            header starts).
        :rtype: tuple(SCPResult, int, bytes, int)
        :raise SpinnmanIOException:
            If there is an error receiving the message
        :raise SpinnmanTimeoutException:
            If there is a timeout before a message is received
        """
        raise NotImplementedError

    @abstractmethod
    def get_scp_data(self, scp_request: AbstractSCPRequest) -> bytes:
        """
        Returns the data of an SCP request as it would be sent down this
        connection.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def chip_x(self) -> int:
        """
        The X-coordinate of the chip at which messages sent down this
        connection will arrive at first.

        :rtype: int
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def chip_y(self) -> int:
        """
        The Y-coordinate of the chip at which messages sent down this
        connection will arrive at first.

        :rtype: int
        """
        raise NotImplementedError



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Dict, List, Tuple
from spinnman.model.enums import P2PTableRoute

_ONE_WORD = struct.Struct("<I")


class P2PTable(object):
    """
    Represents a P2P routing table read from the machine.
    """
    __slots__ = [
        "_height",
        "_routes",
        "_width"]

    def __init__(self, width: int, height: int,
                 column_data: List[Tuple[bytes, int]]):
        """
        :param int width:
        :param int height:
        :param list(tuple(bytes,int)) column_data:
        """
        self._routes: Dict[Tuple[int, int], P2PTableRoute] = dict()
        self._width = width
        self._height = height
        for x, (data, offset) in enumerate(column_data):
            y = 0
            pos = 0
            while y < height:
                next_word, = _ONE_WORD.unpack_from(data, offset + (pos * 4))
                pos += 1
                for entry in range(min(8, height - y)):
                    route = P2PTableRoute((next_word >> (3 * entry)) & 0b111)
                    if route is not P2PTableRoute.NONE:
                        self._routes[x, y] = route
                    y += 1

    @staticmethod
    def get_n_column_bytes(height: int) -> int:
        """
        Get the number of bytes to be read for each column of the table.

        :param int height: The height of the machine
        """
        return ((height + 7) // 8) * 4

    @staticmethod
    def get_column_offset(column: int) -> int:
        """
        Get the offset of the next column in the table from the P2P base
        address.

        :param int column: The column to be read
        """
        return (((256 * column) // 8) * 4)

    @property
    def width(self) -> int:
        """
        The width of the machine that this table represents.

        :rtype: int
        """
        return self._width

    @property
    def height(self) -> int:
        """
        The height of the machine that this table represents.

        :rtype: int
        """
        return self._height

    def iterchips(self):
        """
        Get an iterator of tuples of (x, y) coordinates in the table.

        :rtype: iterable(tuple(int,int))
        """
        return iter(self._routes.keys())

    def is_route(self, x, y):
        """
        Determines if there is a route in the P2P table to the given chip.

        :param int x: The x-coordinate of the chip to look up
        :param int y: The y-coordinate of the chip to look up
        :rtype: bool
        """
        return (
            (x, y) in self._routes and
            self._routes[(x, y)] != P2PTableRoute.NONE)

    def get_route(self, x: int, y: int) -> P2PTableRoute:
        """
        Get the route to follow from this chip to the given chip.

        :param int x: The x-coordinate of the chip to find the route to
        :param int y: The y-coordinate of the chip to find the route to
        :rtype: P2PTableRoute
        """
        return self._routes.get((x, y), P2PTableRoute.NONE)

    @property
    def n_routes(self):
        """ The number of routes in the table

        :rtype: int
        """
        return len(self._routes)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse

_NNP_FORWARD_RETRY = (0x3f << 24) | (0x1E << 16)
_NNP_FLOOD_FILL_START = 6


class FloodFillData(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to start a flood fill of data.
    """
    __slots__ = (
        "_data_to_write",
        "_offset",
        "_size")

    def __init__(
            self, nearest_neighbour_id: int, block_no: int, base_address: int,
            data: bytes, offset: int = 0, length: Optional[int] = None):
        """
        :param int nearest_neighbour_id:
            The ID of the packet, between 0 and 127
        :param int block_no: Which block this block is, between 0 and 255
        :param int base_address:
            The base address where the data is to be loaded
        :param bytes data:
            The data to load, between 4 and 256 bytes and the size must be
            divisible by 4
        """
        # pylint: disable=too-many-arguments
        self._offset = offset
        self._data_to_write = data
        if length is None:
            self._size = len(data)
        else:
            self._size = length

        argument_1 = _NNP_FORWARD_RETRY | nearest_neighbour_id
        argument_2 = (block_no << 16) | (((self._size // 4) - 1) << 8)

        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_FFD),
            argument_1=argument_1, argument_2=argument_2,
            argument_3=base_address, data=None)

    @property
    def bytestring(self) -> bytes:
        datastring = super().bytestring
        data = self._data_to_write[self._offset:self._offset + self._size]
        return datastring + bytes(data)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Flood Fill", "CMD_NNP:NNP_FFS")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.model import VersionInfo
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest, BMPResponse)
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.enums import SCPCommand


class BMPGetVersion(BMPRequest['_BMPVersion']):
    """
    An SCP request to read the version of software running on a core.
    """
    __slots__ = ()

    def __init__(self, board: int):
        """
        :param int board: The board to get the version from
        :raise SpinnmanInvalidParameterException:
            * If the chip coordinates are out of range
            * If the processor is out of range
        """
        super().__init__(
            board, SCPRequestHeader(command=SCPCommand.CMD_VER))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> '_BMPVersion':
        return _BMPVersion("Read BMP version", SCPCommand.CMD_VER)


class _BMPVersion(BMPResponse[VersionInfo]):
    def _parse_payload(self, data: bytes, offset: int) -> VersionInfo:
        return VersionInfo(data, offset)

    @property
    def version_info(self) -> VersionInfo:
        """
        The version information received.

        :rtype: VersionInfo
        """
        return self._value
